#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>
#include "causalito.h"
#include "model.h"
#include "misc.h"
#include "graph.h"
#include "maxflow.h"
#include "globals.h"

/*
 * Search a maximal flow in the graph using the Ford and Fulkerson
 * algorithm.
 *
 * Inputs: the input graph 'g', the name of the source node 'source'
 * and the name of the 'sink' node
 *
 * Output: always return success. The graph given as input is
 * modified: the flow associated with each arc is modified to reflect
 * the computed maximal flow.
 *
 * Warnings: 1) Be sure that there is a backward link from sink to
 * source. 2) The initial flows are not set to zero since you may give
 * a graph with a given flow. BE sURE THAT THE INITIAL FLOWS ARE SET
 * IN A CONSISTANT WAY FOR EACH ARC, INCLUDING THE BACKWARD ARC. IF
 * NOT, YOU MAY GET A SIGSEV, AN INFINITE LOOR OR OTHER WEIRD RESULTS
 * AND/OR CRASH!
 */




/**
 * max_flow:
 * @g: 
 * @source: 
 * @sink: 
 * 
 * 
 * 
 * Return value: 
 **/
int max_flow(Graph g, char * source, char * sink)
{
  int delta;
  
  for (;;) {
    delta = mark_path(g, source, sink); /* search an augmenting path */
/*     debug_print_graph(g); */
    if (!delta) { /* if not found, then exit */ 
      return delta;
    }
    change_flow(g, source, sink, delta); /* else change flow following augmenting path */
  }
  return delta;
}


/**
 * mark_path: Search an augmenting path from source to sink through
 * graph
 * @g: 
 * @source: 
 * @sink: 
 * 
 * 
 * 
 * Return value: 
 **/
int mark_path(Graph g, char * source, char * sink)
{
  int delta;
  Arc backward, u;
  Node sink_node, x, y;
  GSList * N = NULL;

  /* search for the backward arc from sink to source */
  backward = is_arc_in_graph(sink, source, g);
  delta = backward->capacity - backward->flow;
  add_string_in_list(source, N);
  x = is_node_in_graph(source, g);
  x->delta = delta;
  while (!is_string_in_list(sink, N) && delta > 0) {
/*     debug_print_char_list(N); */
    if ((u = search_arc_in_wNplus_with_flow_lower_than_capa(N, g, backward))) {
      x = u->start;
      y = u->end;
      add_string_in_list(y->name, N);
      y->pred = x;
      y->delta = min(x->delta, u->capacity - u->flow);      
    } else {
      if ((u = search_arc_in_wNminus_with_positive_flow(N, g, backward))) {
      x = u->end;
      y = u->start;
      add_string_in_list(y->name, N);
      y->pred = x;
      y->delta = min(x->delta, u->flow);      
      } else {
	delta = 0;
      }
    }
  }
  if (is_string_in_list(sink, N)) {
    sink_node = is_node_in_graph(sink, g);
    delta = sink_node->delta;
  }
  return delta;
}

/**
 * search_arc_in_wNplus_with_flow_lower_than_capa:
 * @N: 
 * @g: 
 * @bw: 
 * 
 * 
 * 
 * Return value: 
 **/
Arc search_arc_in_wNplus_with_flow_lower_than_capa(GSList * N, Graph g, Arc bw)
{
  Arc u = NULL;
  int i, l;

  l = g_slist_length(g->arcs_list);
  for (i = 0 ; i < l; i++) {
    u = g_slist_nth_data(g->arcs_list, i);
    if (u != bw
        && is_string_in_list(u->start->name, N)
	&& !is_string_in_list(u->end->name, N)
	&& u->flow < u->capacity) {
      return u;
    }
    else u = NULL;
  }
  return u;
}

/**
 * search_arc_in_wNminus_with_positive_flow:
 * @N: 
 * @g: 
 * @bw: 
 * 
 * 
 * 
 * Return value: 
 **/
Arc search_arc_in_wNminus_with_positive_flow(GSList * N, Graph g, Arc bw)
{
  Arc u = NULL;
  int i, l;

  l = g_slist_length(g->arcs_list);
  for (i = 0 ; i < l; i++) {
    u = g_slist_nth_data(g->arcs_list, i);
    if (u != bw
  && !is_string_in_list(u->start->name, N)
	&& is_string_in_list(u->end->name, N)
	&& u->flow > 0) {
      return u;
    }
    else u = NULL;
  }
  return u;
}

/**
 * change_flow:
 * @g: 
 * @source: 
 * @sink: 
 * @delta: 
 * 
 * 
 * 
 * Return value: 
 **/
int change_flow(Graph g, char * source, char * sink, int delta)
{
  Node x, y;
  Arc backward, u;

  x = is_node_in_graph(sink, g);
  backward = is_arc_in_graph(sink, source, g);
  set_arc_flow(backward, backward->flow + delta);
  while (strcmp(x->name, source)) {
    y = x->pred;
    u = is_arc_in_graph(y->name, x->name, g);
    if (u) {
      u->flow = u->flow + delta;
    } else {
      u = is_arc_in_graph(x->name, y->name, g);
      u->flow = u->flow - delta;
    }
    x = y;
  }
  return 0;
}

/**
 * debug_print_char_list:
 * @L: 
 * 
 * 
 * 
 * Return value: 
 **/
/* int debug_print_char_list(GSList *L) */
/* { */
/*   int l, i; */

/*   DEBUG_ENTER_FUNC; */
/*   l = g_slist_length(L); */
/*   for (i = 0 ; i < l ; i++) */
/*     printf("%s ", (char *) g_slist_nth_data(L, i)); */
/*   printf("\n"); */
/*   DEBUG_EXIT_FUNC;          */
/*   return 0; */
/* } */

/*** Local Variables: ***/
/*** change-log-default-name: "../ChangeLog" ***/
/*** End: ***/
