#include <stdio.h>
#include <glib.h>
#include <string.h>
#include "readmodel.h"
#include "causalito.h"
#include "model.h"
#include "misc.h"
#include "globals.h"

/*
 * Read and init the model.
 *
 * For each line of the model, get this line, parse lines to get
 * tokens, check that tokens represent a valid equation and if yes add
 * it to the model. Finally check the model
 *
 * WARNING: the checks made are not very robust. So take care the way
 * you write the model !!!
 */



/**
 * read_model:
 * @void: 
 * 
 * 
 * 
 * Return value: 
 **/
int read_model(void)
{
  char tokens[LINESMAX][LINESIZE];
  int i, t, l = 0;
  char line[LINESIZE];
  Equation e = NULL;
  
  while (fgets(line, LINESIZE, options->indesc) != NULL) {
    l += 1;
    if (line[strlen(line) - 1] == '\n') {
      VERBMSG("Reading line %d: %s", l, line);
    } else
      VERBMSG("Reading line %d: %s\n", l, line);
    t = parse_line(line, tokens);
    VERBMSG("%d tokens found on line %d.\n", t, l);
    if (validate_equation(l, t, tokens)) {
      VERBMSG("Creating equation for line %d.\n", l);
      e = alloc_and_init_equation();
      add_comm_in_equation(e, tokens[0]); /* add comment name*/
      add_comp_in_equation(e, tokens[1]); /* add component name*/
      add_eq_in_equation(e, tokens[2]); /* add equation name */
      if (!strlen(tokens[3])) /* if condition is empty then emit warning */
	WARNMSG("Empty condition in equation %s: value is set to 't'.\n",
		e->eq);
      add_cond_in_equation(e, tokens[3]); /* add condition */
      add_varf_in_equation(e, tokens[4]); /* add forced variable */
      for (i = 5 ; i < t ; i++) { /* for each normal variables */
	if (!strlen(tokens[i])) { /* if variable is empty then emit warning */
	  WARNMSG("Empty variable in equation %s.\n", e->eq);
	} else if (is_string_in_list(tokens[i], e->vars)) {
	  /* if variable already appeared in equation, emit a warning
	     (perhaps the user made a typo) */
	  WARNMSG("Variable %s already appears in equations %s.\n",
		  tokens[i], tokens[2]);
	}
	add_var_in_equation(e, tokens[i]);
      } /* end for */
      /* add the forced variable in case it not declared in the normal
	 variables part */
      add_var_in_equation(e, tokens[4]);
      VERBMSG("Adding equation in model.\n");
      model->eqs_list = g_slist_append(model->eqs_list, e);
    } 
  }
  if (!complete_and_validate_model()) {
    return FAILURE;
  }
  return SUCCESS;
}

/*
 * For each line read, search the tokens, i.e. any data surrounded by
 * square brackets and returns the number of tokens
 */


/**
 * parse_line:
 * @: 
 * @: 
 * 
 * 
 * 
 * Return value: 
 **/
int parse_line(char line[LINESIZE], char toks[LINESMAX][LINESIZE])
{
  char c;
  int t = 0, tok_idx = 0, line_idx = 0, tok_found = False;

  while ((c = line[line_idx++]) != '\0') {
    switch (c) {
    case OCHAR:
      if (!tok_found) {
	tok_found = True;
      } else
	toks[t][tok_idx++] = c;
      break;
    case CCHAR:
      if (tok_found) {
	tok_found = False;
	toks[t][tok_idx] = '\0';
	VERBMSG("Token %d found: [%s]\n", t + 1, toks[t]);
	t += 1;
	tok_idx = 0;
      }
      break;
    default:
      if (tok_found)
	toks[t][tok_idx++] = c;
      break;
    } /* end of switch */
  } /* end of while */

  /*
   * We have to fill at least 6 elements of toks, so if the number of
   * token is lower than 6, we complete with empty string since an
   * empty string tells us that there is no corresponding element in
   * the equation. Useful for further tests in validate_equation
   */
  for (tok_idx = t ; tok_idx <= 5 ; tok_idx++)
    toks[tok_idx][0] = '\0';
  return t;
}

/*
 * Check that the equation is correct i.e. there is at least an
 * equation name and one variable (forced or not).
 */




/**
 * validate_equation:
 * @l: 
 * @t: 
 * @: 
 * 
 * 
 * 
 * Return value: 
 **/
int validate_equation(int l, int t, char toks[LINESMAX][LINESIZE])
{
  int  i;
  static char *type[] = {
    "comment", "component", "equation", "condition",
    "forced variable", "normal variable"
  };

  if (t < 5) {
    VERBMSG("Not enough tokens on line %d\n", l);
    WARNMSG("Line %d taken as comments: disregarded.\n", l);
    return FAILURE;
  }
  VERBMSG("Checking tokens found on line %d\n", l);
  for (i = 0; i < t; i++) {
    VERBMSG("...%s%s found\n", strlen(toks[i]) ? "" : "no ",
	    i < 5 ? type[i] : type[5]);
  }
  if (strlen(toks[2]) && (strlen(toks[4]) || strlen(toks[5]))) {
    VERBMSG("Line %d is a valid equation.\n", l);
  } else {
    WARNMSG("Line %d is not a valid equation: disregarded.\n", l);
    VERBMSG("A valid equation need at least an equation name\n");
    VERBMSG("plus a forced variable or one normal variable.\n");
    return FAILURE;
  }
  return SUCCESS;
}

/**
 * complete_and_validate_model:
 * @void: 
 * 
 * 
 * 
 * Return value: 
 **/
int complete_and_validate_model(void)
{
  int i, j, n;
  Equation e;
  char * s, tampon[2048];
  
  VERBMSG("Completing and validating model...\n");
  /* checking number of equations */
  n = g_slist_length(model->eqs_list);
  if (n == 0) {
    ERRMSG("Model does not contain any equation.\n");
    return FAILURE;
  }
  model->neqs = n;
  VERBMSG("%d equations found.\n",  model->neqs);
  for (i = 1 ; i <= model->neqs ; i++) {
    e = g_slist_nth_data(model->eqs_list, i - 1);
    VERBMSG("Equation %s is being treated (number %d).\n", e->eq, i);
    /*
     * List equations names
     */
    if (is_string_in_list(e->eq, model->eqs_name_list)) {
      ERRMSG("There is already an equation named %s in the model.\n", e->eq);
      return FAILURE;
    }
    /* no strdup needed, copy only the pointer */
    model->eqs_name_list = g_slist_append(model->eqs_name_list, e->eq);
    /*
     * List variables names
     */
    e->nvars = g_slist_length(e->vars);
    sprintf(tampon, "Equation %s has %d variables: ", e->eq, e->nvars);
    for (j = 1 ; j <= e->nvars; j++) {
      s = g_slist_nth_data(e->vars, j - 1);
      strcat(tampon, s);
      strcat(tampon, " ");
      if (!is_string_in_list(s, model->vars_name_list))
	model->vars_name_list = g_slist_append(model->vars_name_list, s);
    }
    VERBMSG("%s\n", tampon);
    /*
     * Conditions
     */
    if (strcmp(e->cond, "t")) {
      if (* e->cond == '-') {
	e->positive = False;
	e->depc = strdup(e->cond + 1);
      } else {
	e->depc = strdup(e->cond);
      }
      VERBMSG("Equation %s is submitted to %s condition %s.\n",
	      e->eq, e->positive ? "positive" : "negative", e->depc);
      if (!is_string_in_list(e->depc, model->conds_name_list))
	model->conds_name_list = g_slist_append(model->conds_name_list,
						e->depc);
    } else VERBMSG("Equation %s is always true.\n", e->eq);
  }
  model->nvars = g_slist_length(model->vars_name_list);
  model->nconds = g_slist_length(model->conds_name_list);
  
  /* Summarizing model data */
  OUTMSG("Summarizing model data...\n");
  OUTMSG("%d equations: ", model->neqs);
  for (i = 1 ; i <= model->neqs ; i++) {
    OUTMSG("%s ", (char *) g_slist_nth_data(model->eqs_name_list, i - 1));
  }
  OUTMSG("\n");
  OUTMSG("%d variables: ", model->nvars);
  for (i = 1 ; i <= model->nvars ; i++) {
    OUTMSG("%s ", (char *) g_slist_nth_data(model->vars_name_list, i - 1));
  }
  OUTMSG("\n");
  OUTMSG("%d conditions: ", model->nconds);
  for (i = 1 ; i <= model->nconds ; i++) {
    OUTMSG("%s ", (char *) g_slist_nth_data(model->conds_name_list, i - 1));
  }
  OUTMSG("\n");

  compute_Gray_code(model->conds_name_list);
  OUTMSG("%d operating modes.\n", model->nconfs);
  VERBMSG("List of modes:\n");
  if (model->nconfs == 1) {
    VERBMSG("Only one operating mode (no equations with conditions).\n")
  } else if (model->nconfs >= 2) {
    n = 0;
    for (i = 0 ; i < model->nconfs ; i++) {
      * tampon = '\0';
      for (j = 0 ; j < model->nconds ; j++) {
	sprintf(tampon, "%s %5s", tampon,
		(char *) g_slist_nth_data(model->confs_list, n++));
      }
      VERBMSG("%s\n", tampon);
    }
  } else BUG("model->nconfs = %d\n", model->nconfs);
  return SUCCESS;
}

/*** Local Variables: ***/
/*** change-log-default-name: "../ChangeLog" ***/
/*** End: ***/
